#version 430 core

in vec3 position;
in mat3 tbnMatrix;
in vec2 uvmap;

out vec4 color;

uniform vec3 viewPosition;

// material parameters
uniform sampler2D albedoTexture;
uniform sampler2D normalTexture;
uniform sampler2D rmaTexture;

const float PI = 3.14159265359;

struct DirectionalLight {
	vec3 direction;
	mat3 color;
    mat4 lightSpaceMatrix;
};

struct PointLight {
	vec3 position;
	mat3 color;
	vec3 attenuation;
};

#define MAX_DIRECTIONAL_LIGHTS 3
uniform DirectionalLight directionalLights[MAX_DIRECTIONAL_LIGHTS];
uniform sampler2D shadowMaps[MAX_DIRECTIONAL_LIGHTS];
uniform int numDirectionalLights;
#define MAX_POINT_LIGHTS 16
uniform PointLight pointLights[MAX_POINT_LIGHTS];
uniform int numPointLights;

// The following functions are roughly taken from learnopengl.com
// But altered to remove duplicate computations
float DistributionGGX(float NdotH, float roughness) 
{
    float a = roughness*roughness;
    float a2 = a*a;
    float NdotH2 = NdotH*NdotH;

    float nom   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return nom / max(denom, 0.0000001); // prevent divide by zero for roughness=0.0 and NdotH=1.0
}

// ----------------------------------------------------------------------------
float GeometrySchlickGGX(float NdotV, float k)
{
    float nom   = NdotV;
    float denom = NdotV * (1.0 - k) + k;

    return nom / denom;
}

// ----------------------------------------------------------------------------
float GeometrySmith(float NdotV, float NdotL, float roughness)
{
    
    // Better performance
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;
    float ggx1 = GeometrySchlickGGX(NdotV, k);
    float ggx2 = GeometrySchlickGGX(NdotL, k);

    return ggx1 * ggx2;
}

// ----------------------------------------------------------------------------
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
}

// ----------------------------------------------------------------------------
vec3 calulatePBR(float NdotV, vec3 N, vec3 albedo, vec3 rma, vec3 F0, vec3 V, vec3 Lm, mat3 lightColor) {
    float roughness = rma[0];
    float metallic = rma[1];
    float ao = rma[2];

	vec3 H = normalize(Lm + V);
    float NdotL = dot(N, Lm);
    if(NdotL < 0)
        return vec3(0);

    float HdotV = clamp(dot(H, V), 0.0, 1.0);
    float NdotH = max(dot(N, H), 0.0);

	// Cook-Torrance BRDF
    float NDF = DistributionGGX(NdotH, roughness);   
    float G   = GeometrySmith(NdotV, NdotL, roughness);      
    vec3 F    = fresnelSchlick(HdotV, F0);
    
    // scale light by NdotL

    vec3 nominator    = NDF * G * F; 
    float denominator = 4 * NdotV * NdotL;
    vec3 specular = nominator / max(denominator, 0.00001);
	// kS is equal to Fresnel
    vec3 kS = F;
	vec3 kD = vec3(1.0) - kS;
	kD *= 1.0 - metallic;

    // return outgoing radiance Lo
    return (kD * lightColor[1] * albedo / PI + specular * lightColor[2]) * NdotL;
}

float ShadowCalculation(sampler2D shadowTexture, vec4 fragposlightspace)
{
    // perform perspective divide
    vec3 projCoords = fragposlightspace.xyz / fragposlightspace.w;
    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    //float closestDepth = texture(shadowMap, projCoords.xy).r; 
    // get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
    // check whether current frag pos is in shadow
	
	// apply shadow bias to avoid shadow acne
	float bias = 0.003;
    //float shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;
	// PCF - filter for "smoother" shadows
	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowTexture, 0);
	for(int x = -1; x <= 1; ++x)
	{
		for(int y = -1; y <= 1; ++y)
		{
			float pcfDepth = texture(shadowTexture, projCoords.xy + vec2(x, y) * texelSize).r; 
			shadow += currentDepth - bias > pcfDepth ? 1.0 : 0.0;        
		}    
	}
	shadow /= 9.0;
	
	// for area out of range, to not show it in shadow
	if (projCoords.z > 1.0)
		shadow = 0.0;

    return shadow;
}

void main() {
    vec3 tangentSpaceNormal = texture(normalTexture, uvmap).rgb * 2.0 - 1.0;
	vec3 norm = normalize(tbnMatrix * tangentSpaceNormal);

    vec3 albedo = pow(texture(albedoTexture, uvmap).rgb, vec3(2.2, 2.2, 2.2));
    vec3 rma = texture(rmaTexture, uvmap).rgb;
    float metallic = rma[1];
    float ao = rma[2];

	// Direction to camera
	vec3 V = normalize(viewPosition-position);
	float NdotV = max(dot(norm, V), 0.0);

	vec3 F0 = mix(vec3(0.04), albedo, metallic);
	// reflectance equation
    vec3 Lo = vec3(0.0);

    for(int i = 0; i < numDirectionalLights; i++) {
        DirectionalLight light = directionalLights[i];

        Lo += calulatePBR(NdotV, norm, albedo, rma, F0, V, -normalize(light.direction), light.color);
    }
       
    for(int i = 0; i < numDirectionalLights; i++) {
        DirectionalLight light = directionalLights[i];
        float shadow = ShadowCalculation(shadowMaps[i], light.lightSpaceMatrix * vec4(position, 1.0));
        Lo *= (1.0 - shadow);
    }

	for(int i = 0; i < numPointLights; i++) {
		PointLight light = pointLights[i];
		float d = distance(light.position, position);
		float Fatt = 1 / (light.attenuation.x + light.attenuation.y*d + light.attenuation.z*d*d);

		vec3 Lm = normalize(light.position - position);
				
        Lo += calulatePBR(NdotV, norm, albedo, rma, F0, V, Lm, light.color * Fatt);
    }
	
	vec3 ambient = vec3(0.03) * ao * albedo;
    color = vec4(ambient + Lo, 1.0);
}